@@ -1,26 +1,29 @@
 from __future__ import unicode_literals
+
 import gzip
 import mimetypes
 import os
 import posixpath
 import shutil
-from io import TextIOBase
 from contextlib import closing
 from functools import wraps
+from io import TextIOBase
 from tempfile import SpooledTemporaryFile
+from threading import local
+from urllib.parse import urljoin, urlsplit, urlunsplit
+
 import boto3
-from botocore.exceptions import ClientError
 from botocore.client import Config
+from botocore.exceptions import ClientError
 from django.conf import settings
-from django.core.exceptions import ImproperlyConfigured
-from django.core.files.storage import Storage
-from django.core.files.base import File
-from django.core.signals import setting_changed
 from django.contrib.staticfiles.storage import ManifestFilesMixin
-from django.utils.six.moves.urllib.parse import urlsplit, urlunsplit, urljoin
+from django.core.exceptions import ImproperlyConfigured
+from django.core.files.base import File
+from django.core.files.storage import Storage
+from django.core.signals import setting_changed
 from django.utils.deconstruct import deconstructible
-from django.utils.encoding import force_bytes, filepath_to_uri, force_text, force_str
-from django.utils.timezone import make_naive
+from django.utils.encoding import filepath_to_uri, force_bytes, force_str, force_text
+from django.utils.timezone import make_naive, utc
 
 
 def _wrap_errors(func):
@@ -29,7 +32,11 @@ def _wrap_errors(func):
         try:
             return func(self, name, *args, **kwargs)
         except ClientError as ex:
-            raise IOError("S3Storage error at {!r}: {}".format(name, force_text(ex)))
+            code = ex.response.get("Error", {}).get("Code", "Unknown")
+            err_cls = OSError
+            if code == "NoSuchKey":
+                err_cls = FileNotFoundError
+            raise err_cls("S3Storage error at {!r}: {}".format(name, force_text(ex)))
     return _do_wrap_errors
 
 
@@ -41,6 +48,34 @@ def _temporary_file():
     return SpooledTemporaryFile(max_size=1024*1024*10)  # 10 MB.
 
 
+def _to_sys_path(name):
+    return name.replace("/", os.sep)
+
+
+def _to_posix_path(name):
+    return name.replace(os.sep, "/")
+
+
+def _wrap_path_impl(func):
+    @wraps(func)
+    def do_wrap_path_impl(self, name, *args, **kwargs):
+        # The default implementations of most storage methods assume that system-native paths are used. But we deal with
+        # posix paths. We fix this by converting paths to system form, passing them to the default implementation, then
+        # converting them back to posix paths.
+        return _to_posix_path(func(self, _to_sys_path(name), *args, **kwargs))
+    return do_wrap_path_impl
+
+
+def unpickle_helper(cls, kwargs):
+    return cls(**kwargs)
+
+
+Settings = type(force_str("Settings"), (), {})
+
+
+_UNCOMPRESSED_SIZE_META_KEY = "uncompressed_size"
+
+
 class S3File(File):
 
     """
@@ -57,6 +92,34 @@ class S3File(File):
         return super(S3File, self).open(mode)
 
 
+class _Local(local):
+
+    """
+    Thread-local connection manager.
+
+    Boto3 objects are not thread-safe.
+    http://boto3.readthedocs.io/en/latest/guide/resources.html#multithreading-multiprocessing
+    """
+
+    def __init__(self, storage):
+        connection_kwargs = {
+            "region_name": storage.settings.AWS_REGION,
+        }
+        if storage.settings.AWS_ACCESS_KEY_ID:
+            connection_kwargs["aws_access_key_id"] = storage.settings.AWS_ACCESS_KEY_ID
+        if storage.settings.AWS_SECRET_ACCESS_KEY:
+            connection_kwargs["aws_secret_access_key"] = storage.settings.AWS_SECRET_ACCESS_KEY
+        if storage.settings.AWS_SESSION_TOKEN:
+            connection_kwargs["aws_session_token"] = storage.settings.AWS_SESSION_TOKEN
+        if storage.settings.AWS_S3_ENDPOINT_URL:
+            connection_kwargs["endpoint_url"] = storage.settings.AWS_S3_ENDPOINT_URL
+        self.session = boto3.session.Session()
+        self.s3_connection = self.session.client("s3", config=Config(
+            s3={"addressing_style": storage.settings.AWS_S3_ADDRESSING_STYLE},
+            signature_version=storage.settings.AWS_S3_SIGNATURE_VERSION,
+        ), **connection_kwargs)
+
+
 @deconstructible
 class S3Storage(Storage):
 
@@ -68,6 +131,7 @@ class S3Storage(Storage):
         "AWS_REGION": "us-east-1",
         "AWS_ACCESS_KEY_ID": "",
         "AWS_SECRET_ACCESS_KEY": "",
+        "AWS_SESSION_TOKEN": "",
     }
 
     default_s3_settings = {
@@ -83,13 +147,16 @@ class S3Storage(Storage):
         "AWS_S3_CONTENT_LANGUAGE": "",
         "AWS_S3_METADATA": {},
         "AWS_S3_ENCRYPT_KEY": False,
+        "AWS_S3_KMS_ENCRYPTION_KEY_ID": "",
         "AWS_S3_GZIP": True,
+        "AWS_S3_SIGNATURE_VERSION": "s3v4",
+        "AWS_S3_FILE_OVERWRITE": False
     }
 
     s3_settings_suffix = ""
 
     def _setup(self):
-        self.settings = type(force_str("Settings"), (), {})()
+        self.settings = Settings()
         # Configure own settings.
         for setting_key, setting_default_value in self.default_auth_settings.items():
             setattr(
@@ -110,26 +177,31 @@ class S3Storage(Storage):
                 ),
             )
         # Validate settings.
+        if not self.settings.AWS_S3_BUCKET_NAME:
+            raise ImproperlyConfigured(f"Setting AWS_S3_BUCKET_NAME{self.s3_settings_suffix} is required.")
+        if not (
+            (self.settings.AWS_ACCESS_KEY_ID and self.settings.AWS_SECRET_ACCESS_KEY) or
+            self.settings.AWS_SESSION_TOKEN
+        ):
+            raise ImproperlyConfigured(
+                f"Settings AWS_ACCESS_KEY_ID{self.s3_settings_suffix} "
+                f"and AWS_SECRET_ACCESS_KEY{self.s3_settings_suffix}, or "
+                f"AWS_SESSION_TOKEN{self.s3_settings_suffix}, is required."
+            )
         if self.settings.AWS_S3_PUBLIC_URL and self.settings.AWS_S3_BUCKET_AUTH:
-            raise ImproperlyConfigured("Cannot use AWS_S3_BUCKET_AUTH with AWS_S3_PUBLIC_URL.")
-        # Connect to S3.
-        connection_kwargs = {
-            "region_name": self.settings.AWS_REGION,
-        }
-        if self.settings.AWS_ACCESS_KEY_ID:
-            connection_kwargs["aws_access_key_id"] = self.settings.AWS_ACCESS_KEY_ID
-        if self.settings.AWS_SECRET_ACCESS_KEY:
-            connection_kwargs["aws_secret_access_key"] = self.settings.AWS_SECRET_ACCESS_KEY
-        if self.settings.AWS_S3_ENDPOINT_URL:
-            connection_kwargs["endpoint_url"] = self.settings.AWS_S3_ENDPOINT_URL
-        self.s3_connection = boto3.client("s3", config=Config(
-            s3={"addressing_style": self.settings.AWS_S3_ADDRESSING_STYLE},
-        ), **connection_kwargs)
+            raise ImproperlyConfigured(
+                f"Cannot use AWS_S3_BUCKET_AUTH{self.s3_settings_suffix} "
+                f"with AWS_S3_PUBLIC_URL{self.s3_settings_suffix}."
+            )
+        # Create a thread-local connection manager.
+        self._connections = _Local(self)
+
+    @property
+    def s3_connection(self):
+        return self._connections.s3_connection
 
     def _setting_changed_received(self, setting, **kwargs):
         if setting.startswith("AWS_"):
-            # HACK: No supported way to close the HTTP session from boto3... :S
-            self.s3_connection._endpoint.http_session.close()
             self._setup()
 
     def __init__(self, **kwargs):
@@ -148,12 +220,15 @@ class S3Storage(Storage):
         # All done!
         super(S3Storage, self).__init__()
 
+    def __reduce__(self):
+        return unpickle_helper, (self.__class__, self._kwargs)
+
     # Helpers.
 
     def _get_key_name(self, name):
         if name.startswith("/"):
             name = name[1:]
-        return posixpath.normpath(posixpath.join(self.settings.AWS_S3_KEY_PREFIX, name.replace(os.sep, "/")))
+        return posixpath.normpath(posixpath.join(self.settings.AWS_S3_KEY_PREFIX, _to_posix_path(name)))
 
     def _object_params(self, name):
         params = {
@@ -187,8 +262,13 @@ class S3Storage(Storage):
         if content_langauge:
             params["ContentLanguage"] = content_langauge
         # Set server-side encryption.
-        if self.settings.AWS_S3_ENCRYPT_KEY:
-            params["ServerSideEncryption"] = "AES256"
+        if self.settings.AWS_S3_ENCRYPT_KEY:  # If this if False / None / empty then no encryption
+            if isinstance(self.settings.AWS_S3_ENCRYPT_KEY, str):
+                params["ServerSideEncryption"] = self.settings.AWS_S3_ENCRYPT_KEY
+                if self.settings.AWS_S3_KMS_ENCRYPTION_KEY_ID:
+                    params["SSEKMSKeyId"] = self.settings.AWS_S3_KMS_ENCRYPTION_KEY_ID
+            else:
+                params["ServerSideEncryption"] = "AES256"
         # All done!
         return params
 
@@ -239,14 +319,16 @@ class S3Storage(Storage):
                 with closing(gzip.GzipFile(name, "wb", 9, temp_file)) as gzip_file:
                     shutil.copyfileobj(content, gzip_file)
                 # Only use the compressed version if the zipped version is actually smaller!
-                if temp_file.tell() < content.tell():
+                orig_size = content.tell()
+                if temp_file.tell() < orig_size:
                     temp_file.seek(0)
                     content = temp_file
                     put_params["ContentEncoding"] = "gzip"
+                    put_params["Metadata"][_UNCOMPRESSED_SIZE_META_KEY] = "{:d}".format(orig_size)
                 else:
                     content.seek(0)
         # Save the file.
-        self.s3_connection.put_object(Body=content, **put_params)
+        self.s3_connection.put_object(Body=content.read(), **put_params)
         # Close all temp files.
         for temp_file in temp_files:
             temp_file.close()
@@ -255,6 +337,20 @@ class S3Storage(Storage):
 
     # Subsiduary storage methods.
 
+    @_wrap_path_impl
+    def get_valid_name(self, name):
+        return super(S3Storage, self).get_valid_name(name)
+
+    @_wrap_path_impl
+    def get_available_name(self, name, max_length=None):
+        if self.settings.AWS_S3_FILE_OVERWRITE:
+            return _to_posix_path(name)
+        return super(S3Storage, self).get_available_name(name, max_length=max_length)
+
+    @_wrap_path_impl
+    def generate_filename(self, filename):
+        return super(S3Storage, self).generate_filename(filename)
+
     @_wrap_errors
     def meta(self, name):
         """Returns a dictionary of metadata associated with the key."""
@@ -265,17 +361,28 @@ class S3Storage(Storage):
         self.s3_connection.delete_object(**self._object_params(name))
 
     def exists(self, name):
-        # We also need to check for directory existence, so we'll list matching
-        # keys and return success if any match.
-        results = self.s3_connection.list_objects_v2(
-            Bucket=self.settings.AWS_S3_BUCKET_NAME,
-            MaxKeys=1,
-            Prefix=self._get_key_name(name),
-        )
-        return bool(results["KeyCount"])
+        name = _to_posix_path(name)
+        if name.endswith("/"):
+            # This looks like a directory, but on S3 directories are virtual, so we need to see if the key starts
+            # with this prefix.
+            results = self.s3_connection.list_objects_v2(
+                Bucket=self.settings.AWS_S3_BUCKET_NAME,
+                MaxKeys=1,
+                Prefix=self._get_key_name(name) + "/",  # Add the slash again, since _get_key_name removes it.
+            )
+            return "Contents" in results
+        # This may be a file or a directory. Check if getting the file metadata throws an error.
+        try:
+            self.meta(name)
+        except OSError:
+            # It's not a file, but it might be a directory. Check again that it's not a directory.
+            return self.exists(name + "/")
+        else:
+            return True
 
     def listdir(self, path):
-        path = self._get_key_name(path) + "/"
+        path = self._get_key_name(path)
+        path = "" if path == "." else path + "/"
         # Look through the paths, parsing out directories and paths.
         files = []
         dirs = []
@@ -294,16 +401,28 @@ class S3Storage(Storage):
         return dirs, files
 
     def size(self, name):
-        return self.meta(name)["ContentLength"]
+        meta = self.meta(name)
+        try:
+            if meta["ContentEncoding"] == "gzip":
+                return int(meta["Metadata"]["uncompressed_size"])
+        except KeyError:
+            return meta["ContentLength"]
 
-    def url(self, name):
+    def url(self, name, extra_params=None):
         # Use a public URL, if specified.
         if self.settings.AWS_S3_PUBLIC_URL:
+            if extra_params:
+                raise ValueError(
+                    "Use of extra_params to generate custom URLs is not allowed "
+                    "with AWS_S3_PUBLIC_URL"
+                )
             return urljoin(self.settings.AWS_S3_PUBLIC_URL, filepath_to_uri(name))
         # Otherwise, generate the URL.
+        params = extra_params.copy() if extra_params else {}
+        params.update(self._object_params(name))
         url = self.s3_connection.generate_presigned_url(
             ClientMethod="get_object",
-            Params=self._object_params(name),
+            Params=params,
             ExpiresIn=self.settings.AWS_S3_MAX_AGE_SECONDS,
         )
         # Strip off the query params if we're not interested in bucket auth.
@@ -312,14 +431,16 @@ class S3Storage(Storage):
         # All done!
         return url
 
-    def accessed_time(self, name):
-        return self.modified_time(name)
-
-    def created_time(self, name):
-        return self.modified_time(name)
-
     def modified_time(self, name):
-        return make_naive(self.meta(name)["LastModified"])
+        return make_naive(self.meta(name)["LastModified"], utc)
+
+    created_time = accessed_time = modified_time
+
+    def get_modified_time(self, name):
+        timestamp = self.meta(name)["LastModified"]
+        return timestamp if settings.USE_TZ else make_naive(timestamp)
+
+    get_created_time = get_accessed_time = get_modified_time
 
     def sync_meta_iter(self):
         paginator = self.s3_connection.get_paginator("list_objects_v2")
@@ -330,12 +451,24 @@ class S3Storage(Storage):
         for page in pages:
             for entry in page.get("Contents", ()):
                 name = posixpath.relpath(entry["Key"], self.settings.AWS_S3_KEY_PREFIX)
-                obj = self.meta(name)
+                try:
+                    obj = self.meta(name)
+                except OSError:
+                    # This may be caused by a race condition, with the entry being deleted before it was accessed.
+                    # Alternatively, the key may be something that, when normalized, has a different path, which will
+                    # mean that the key's meta cannot be accessed.
+                    continue
                 put_params = self._object_put_params(name)
                 # Set content encoding.
                 content_encoding = obj.get("ContentEncoding")
                 if content_encoding:
                     put_params["ContentEncoding"] = content_encoding
+                    if content_encoding == "gzip":
+                        try:
+                            put_params["Metadata"][_UNCOMPRESSED_SIZE_META_KEY] = \
+                                obj["Metadata"][_UNCOMPRESSED_SIZE_META_KEY]
+                        except KeyError:
+                            pass
                 # Update the metadata.
                 self.s3_connection.copy_object(
                     ContentType=obj["ContentType"],
@@ -362,7 +495,6 @@ class StaticS3Storage(S3Storage):
     default_s3_settings = S3Storage.default_s3_settings.copy()
     default_s3_settings.update({
         "AWS_S3_BUCKET_AUTH": False,
-        "AWS_S3_MAX_AGE_SECONDS": 60 * 60 * 24 * 365,  # 1 year.
     })
 
     s3_settings_suffix = "_STATIC"
@@ -370,4 +502,16 @@ class StaticS3Storage(S3Storage):
 
 class ManifestStaticS3Storage(ManifestFilesMixin, StaticS3Storage):
 
-    pass
+    default_s3_settings = StaticS3Storage.default_s3_settings.copy()
+    default_s3_settings.update({
+        "AWS_S3_MAX_AGE_SECONDS_CACHED": 60 * 60 * 24 * 365,  # 1 year.
+    })
+
+    def post_process(self, *args, **kwargs):
+        initial_aws_s3_max_age_seconds = self.settings.AWS_S3_MAX_AGE_SECONDS
+        self.settings.AWS_S3_MAX_AGE_SECONDS = self.settings.AWS_S3_MAX_AGE_SECONDS_CACHED
+        try:
+            for r in super(ManifestStaticS3Storage, self).post_process(*args, **kwargs):
+                yield r
+        finally:
+            self.settings.AWS_S3_MAX_AGE_SECONDS = initial_aws_s3_max_age_seconds
